@@ -34,6 +34,10 @@
 *     [13] Cortes, David.
 *          "Isolation forests: looking beyond tree depth."
 *          arXiv preprint arXiv:2111.11639 (2021).
+*     [14] Ting, Kai Ming, Yue Zhu, and Zhi-Hua Zhou.
+*          "Isolation kernel and its effect on SVM"
+*          Proceedings of the 24th ACM SIGKDD
+*          International Conference on Knowledge Discovery & Data Mining. 2018.
 * 
 *     BSD 2-Clause License
 *     Copyright (c) 2019-2022, David Cortes
@@ -108,14 +112,85 @@
 *       Hyperparameters related to imputation might differ between 'imputer' and 'iother' ('imputer' will preserve its
 *       hyperparameters after the merge).
 *       Pass NULL if this is not to be used.
+* - indexer (in, out)
+*       Pointer to indexer object which has already been fit through 'fit_iforest' along with
+*       either 'model' or 'ext_model' in the same call to 'fit_iforest' or through another specialized function.
+*       The imputation nodes from 'ind_other' will be merged into this (will be at the end of vector member 'indices').
+*       Reference points should not differ between 'indexer' and 'ind_other'.
+*       Pass NULL if this is not to be used.
+* - ind_other
+*       Pointer to indexer object which has already been fit through 'fit_iforest' along with
+*       either 'model' or 'ext_model' in the same call to 'fit_iforest' or through another specialized function.
+*       The imputation nodes from this object will be added into 'imputer' (this object will not be modified).
+*       Reference points should not differ between 'indexer' and 'ind_other'.
+*       Pass NULL if this is not to be used.
 */
 void merge_models(IsoForest*     model,      IsoForest*     other,
                   ExtIsoForest*  ext_model,  ExtIsoForest*  ext_other,
-                  Imputer*       imputer,    Imputer*       iother)
+                  Imputer*       imputer,    Imputer*       iother,
+                  TreesIndexer*  indexer,    TreesIndexer*  ind_other)
 {
     size_t curr_size_model = (model != NULL)? (model->trees.size()) : 0;
     size_t curr_size_model_ext = (ext_model != NULL)? (ext_model->hplanes.size()) : 0;
     size_t curr_size_imputer = (imputer != NULL)? (imputer->imputer_tree.size()) : 0;
+    size_t curr_size_indexer = (indexer != NULL)? (indexer->indices.size()) : 0;
+
+    if (imputer != NULL && iother == NULL)
+        throw std::runtime_error("Model to append trees to has imputer, but model to take trees from doesn't.\n");
+    if (indexer != NULL && ind_other == NULL)
+        throw std::runtime_error("Model to append trees to has indexer, but model to take trees from doesn't.\n");
+    if (indexer != NULL && ind_other != NULL)
+    {
+        bool indexer_is_empty = indexer->indices.empty();
+        bool ind_other_is_empty = ind_other->indices.empty();
+        bool model_is_empty = (model != NULL && model->trees.empty()) || (ext_model != NULL && ext_model->hplanes.empty());
+        bool other_is_empty = (other != NULL && other->trees.empty()) || (ext_model != NULL && ext_other->hplanes.empty());
+
+        if (indexer_is_empty && !model_is_empty && ind_other_is_empty && !other_is_empty) {
+            indexer = NULL;
+            ind_other = NULL;
+            goto skip_indexers;
+        }
+
+        if (!model_is_empty && !indexer_is_empty && !other_is_empty && ind_other_is_empty)
+            throw std::runtime_error("Model to append trees to has indexer, but model to take trees from doesn't.\n");
+        if (!model_is_empty && indexer_is_empty && !other_is_empty && !ind_other_is_empty)
+            throw std::runtime_error("Model to take trees from has indexer, but model to append trees to doesn't.\n");
+
+        if (
+            !indexer_is_empty && !ind_other_is_empty &&
+            indexer->indices.front().reference_points.size() != ind_other->indices.front().reference_points.size()
+        ) {
+            throw std::runtime_error("Model to append trees to and model to take trees from have different number of reference points.\n");
+        }
+
+
+        if (
+            !indexer_is_empty &&
+            !ind_other_is_empty &&
+            !indexer->indices.front().node_distances.empty() &&
+            ind_other->indices.front().node_distances.empty()
+        ) {
+            throw std::runtime_error("Model to append trees to has indexer with distances, but model to take trees from has indexer without distances.\n");
+        }
+        if (
+            !indexer_is_empty &&
+            !ind_other_is_empty &&
+            !indexer->indices.front().reference_points.empty() &&
+            ind_other->indices.front().reference_points.empty()
+        ) {
+            throw std::runtime_error("Model to append trees to has indexer with reference points, but model to take trees from has indexer without reference points.\n");
+        }
+        if (
+            !indexer_is_empty &&
+            !ind_other_is_empty &&
+            !indexer->indices.front().reference_indptr.empty() &&
+            ind_other->indices.front().reference_indptr.empty()
+        ) {
+            throw std::runtime_error("Model to append trees to has indexer with kernel reference points, but model to take trees from has indexer without kernel reference points.\n");
+        }
+    }
+    skip_indexers:
 
     try
     {
@@ -124,7 +199,7 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
             if (model == other)
             {
                 auto other_copy = *other;
-                merge_models(model, &other_copy, ext_model, ext_other, imputer, iother);
+                merge_models(model, &other_copy, NULL, NULL, NULL, NULL, NULL, NULL);
                 return;
             }
             model->trees.insert(model->trees.end(),
@@ -138,7 +213,7 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
             if (ext_model == ext_other)
             {
                 auto other_copy = *ext_other;
-                merge_models(model, other, ext_model, &other_copy, imputer, iother);
+                merge_models(NULL, NULL, ext_model, &other_copy, NULL, NULL, NULL, NULL);
                 return;
             }
             ext_model->hplanes.insert(ext_model->hplanes.end(),
@@ -151,20 +226,34 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
             if (imputer == iother)
             {
                 auto other_copy = *iother;
-                merge_models(model, other, ext_model, ext_other, imputer, &other_copy);
+                merge_models(NULL, NULL, NULL, NULL, imputer, &other_copy, NULL, NULL);
                 return;
             }
             imputer->imputer_tree.insert(imputer->imputer_tree.end(),
                                          iother->imputer_tree.begin(),
                                          iother->imputer_tree.end());
         }
+
+        if (indexer != NULL && ind_other != NULL)
+        {
+            if (indexer == ind_other)
+            {
+                auto other_copy = *ind_other;
+                merge_models(NULL, NULL, NULL, NULL, NULL, NULL, indexer, &other_copy);
+                return;
+            }
+            indexer->indices.insert(indexer->indices.end(),
+                                    ind_other->indices.begin(),
+                                    ind_other->indices.end());
+        }
     }
 
-    catch(...)
+    catch (...)
     {
         if (model != NULL) model->trees.resize(curr_size_model);
         if (ext_model != NULL) ext_model->hplanes.resize(curr_size_model_ext);
         if (imputer != NULL) imputer->imputer_tree.resize(curr_size_imputer);
+        if (indexer != NULL) indexer->indices.resize(curr_size_indexer);
         throw;
     }
 }
